"""Probabilistic autoregressive model."""

import logging

import numpy as np
import pandas as pd
import torch
from tqdm import tqdm

from deepecho.models.base import DeepEcho

LOGGER = logging.getLogger(__name__)


class PARNet(torch.nn.Module):
    """PARModel ANN model."""

    def __init__(self, data_size, context_size, hidden_size=32):
        super(PARNet, self).__init__()
        self.context_size = context_size
        self.down = torch.nn.Linear(data_size + context_size, hidden_size)
        self.rnn = torch.nn.GRU(hidden_size, hidden_size)
        self.up = torch.nn.Linear(hidden_size, data_size)

    def forward(self, x, c):
        """Forward passing computation."""
        if isinstance(x, torch.nn.utils.rnn.PackedSequence):
            x, lengths = torch.nn.utils.rnn.pad_packed_sequence(x)
            if self.context_size:
                x = torch.cat(
                    [
                        x,
                        c.unsqueeze(0).expand(x.shape[0], c.shape[0], c.shape[1]),
                    ],
                    dim=2,
                )

            x = self.down(x)
            x = torch.nn.utils.rnn.pack_padded_sequence(x, lengths, enforce_sorted=False)
            x, _ = self.rnn(x)
            x, lengths = torch.nn.utils.rnn.pad_packed_sequence(x)
            x = self.up(x)
            x = torch.nn.utils.rnn.pack_padded_sequence(x, lengths, enforce_sorted=False)

        else:
            if self.context_size:
                x = torch.cat(
                    [
                        x,
                        c.unsqueeze(0).expand(x.shape[0], c.shape[0], c.shape[1]),
                    ],
                    dim=2,
                )

            x = self.down(x)
            x, _ = self.rnn(x)
            x = self.up(x)

        return x


class PARModel(DeepEcho):
    """Probabilistic autoregressive model.

    1. Analyze data types.
        - Categorical/Ordinal: Create mapping.
        - Continuous/Timestamp: Normalize, etc.
        - Count: Compute min and range
    2. Data -> Tensor
        - Categorical/Ordinal: Create one-hot
        - Continuous/Timestamp: Copy into `mu` after normalize, set `var=0.0`
        - Count: Subtract min, divide by range, copy into `r`, set `p=0.0`
    3. Loss
        - Categorical/Ordinal: Cross entropy
        - Continuous/Timestamp: Gaussian likelihood
        - Count: Negative binomial (multiply param + value by range), evaluate loss.
    4. Sample (Tensor -> Tensor)
        - Categorical/Ordinal: Categorical distribution, store as one hot
        - Continuous/Timestamp: Gaussian sample, store into `mu`
        - Count: Negative binomial sample (multiply `r` by range?), divide by range
    5. Tensor -> Data
        - Categorical/Ordinal: Find the maximum value
        - Continuous/Timestamp: Rescale the value in `mu`
        - Count: Multiply by range, add min.

    Args:
        epochs (int):
            The number of epochs to train for. Defaults to 128.
        sample_size (int):
            The number of times to sample (before choosing and
            returning the sample which maximizes the likelihood).
            Defaults to 1.
        cuda (bool):
            Whether to attempt to use cuda for GPU computation.
            If this is False or CUDA is not available, CPU will be used.
            Defaults to ``True``.
        verbose (bool):
            Whether to print progress to console or not.
    """

    def __init__(self, epochs=128, sample_size=1, cuda=True, verbose=True):
        self.epochs = epochs
        self.sample_size = sample_size

        if not cuda or not torch.cuda.is_available():
            device = 'cpu'
        elif isinstance(cuda, str):
            device = cuda
        else:
            device = 'cuda'

        self.device = torch.device(device)
        self.verbose = verbose
        self.loss_values = pd.DataFrame(columns=['Epoch', 'Loss'])

        LOGGER.info('%s instance created', self)

    def __repr__(self):
        """Return a representation of the class object."""
        return (
            f'{self.__class__.__name__}(epochs={self.epochs}, sample_size={self.sample_size},'
            f"cuda='{self.device}', verbose={self.verbose})"
        )

    def _idx_map(self, x, t):
        idx = 0
        idx_map = {}
        for i, t in enumerate(t):
            if t == 'continuous' or t == 'datetime':
                idx_map[i] = {
                    'type': t,
                    'mu': np.nanmean(x[i]),
                    'std': np.nanstd(x[i]),
                    'nulls': pd.isna(x[i]).any(),
                    'indices': (idx, idx + 1, idx + 2),
                }
                idx += 3

            elif t == 'count':
                idx_map[i] = {
                    'type': t,
                    'min': np.nanmin(x[i]),
                    'range': np.nanmax(x[i]) - np.nanmin(x[i]),
                    'nulls': pd.isna(x[i]).any(),
                    'indices': (idx, idx + 1, idx + 2),
                }
                idx += 3

            elif t == 'categorical' or t == 'ordinal':
                idx_map[i] = {'type': t, 'indices': {}}
                idx += 1
                for v in set(x[i]):
                    if pd.isna(v):
                        v = None

                    idx_map[i]['indices'][v] = idx
                    idx += 1

            else:
                raise ValueError(f'Unsupported type: {t}')

        return idx_map, idx

    def _build(self, sequences, context_types, data_types):
        contexts = [[] for _ in range(len(context_types))]
        data = [[] for _ in range(len(data_types))]
        min_length = np.inf
        max_length = -np.inf
        for sequence in sequences:
            sequence_data = sequence['data']
            sequence_context = sequence['context']
            sequence_length = len(sequence_data[0])
            min_length = min(min_length, sequence_length)
            max_length = max(max_length, sequence_length)

            for i in range(len(context_types)):
                contexts[i].append(sequence_context[i])
            for i in range(len(data_types)):
                data[i].extend(sequence_data[i])

        self._fixed_length = min_length == max_length
        self._min_length = min_length
        self._max_length = max_length

        self._ctx_map, self._ctx_dims = self._idx_map(contexts, context_types)
        self._data_map, self._data_dims = self._idx_map(data, data_types)
        self._data_map['<TOKEN>'] = {
            'type': 'categorical',
            'indices': {
                '<START>': self._data_dims,
                '<END>': self._data_dims + 1,
                '<BODY>': self._data_dims + 2,
            },
        }
        self._data_dims += 3

    def _data_to_tensor(self, data):
        seq_len = len(data[0])
        X = []

        x = torch.zeros(self._data_dims)
        x[self._data_map['<TOKEN>']['indices']['<START>']] = 1.0
        X.append(x)

        for i in range(seq_len):
            x = torch.zeros(self._data_dims)
            for key, props in self._data_map.items():
                if key == '<TOKEN>':
                    x[self._data_map['<TOKEN>']['indices']['<BODY>']] = 1.0

                elif props['type'] in ['continuous', 'timestamp']:
                    mu_idx, sigma_idx, missing_idx = props['indices']
                    if pd.isna(data[key][i]) or props['std'] == 0:
                        x[mu_idx] = 0.0
                    else:
                        x[mu_idx] = (data[key][i] - props['mu']) / props['std']

                    x[sigma_idx] = 0.0
                    x[missing_idx] = 1.0 if pd.isna(data[key][i]) else 0.0

                elif props['type'] in ['count']:
                    r_idx, p_idx, missing_idx = props['indices']
                    if pd.isna(data[key][i]) or props['range'] == 0:
                        x[r_idx] = 0.0
                    else:
                        x[r_idx] = (data[key][i] - props['min']) / props['range']

                    x[p_idx] = 0.0
                    x[missing_idx] = 1.0 if pd.isna(data[key][i]) else 0.0

                elif props['type'] in [
                    'categorical',
                    'ordinal',
                ]:  # categorical
                    value = data[key][i]
                    if pd.isna(value):
                        value = None
                    x[props['indices'][value]] = 1.0

                else:
                    raise ValueError()

            X.append(x)

        x = torch.zeros(self._data_dims)
        x[self._data_map['<TOKEN>']['indices']['<END>']] = 1.0
        X.append(x)

        return torch.stack(X, dim=0).to(self.device)

    def _context_to_tensor(self, context):
        if not self._ctx_dims:
            return None

        x = torch.zeros(self._ctx_dims)
        for key, props in self._ctx_map.items():
            if props['type'] in ['continuous', 'datetime']:
                mu_idx, sigma_idx, missing_idx = props['indices']
                x[mu_idx] = (
                    0.0
                    if (pd.isna(context[key]) or props['std'] == 0)
                    else (context[key] - props['mu']) / props['std']
                )
                x[sigma_idx] = 0.0
                x[missing_idx] = 1.0 if pd.isna(context[key]) else 0.0

            elif props['type'] in ['count']:
                r_idx, p_idx, missing_idx = props['indices']
                x[r_idx] = (
                    0.0
                    if (pd.isna(context[key]) or props['range'] == 0)
                    else (context[key] - props['min']) / props['range']
                )
                x[p_idx] = 0.0
                x[missing_idx] = 1.0 if pd.isna(context[key]) else 0.0

            elif props['type'] in ['categorical', 'ordinal']:
                value = context[key]
                if pd.isna(value):
                    value = None
                x[props['indices'][value]] = 1.0

            else:
                raise ValueError()

        return x.to(self.device)

    def fit_sequences(self, sequences, context_types, data_types):
        """Fit a model to the specified sequences.

        Args:
            sequences (list):
                List of sequences. Each sequence is a single training example
                (i.e. an example of a multivariate time series with some context).
                For example, a sequence might look something like::

                    {
                        'context': [1],
                        'data': [
                            [1, 3, 4, 5, 11, 3, 4],
                            [2, 2, 3, 4, 5, 1, 2],
                            [1, 3, 4, 5, 2, 3, 1],
                        ],
                    }

                The "context" attribute maps to a list of variables which
                should be used for conditioning. These are variables which
                do not change over time.

                The "data" attribute contains a list of lists corrsponding
                to the actual time series data such that `data[i][j]` contains
                the value at the jth time step of the ith channel of the
                multivariate time series.
            context_types (list):
                List of strings indicating the type of each value in context.
                he value at `context[i]` must match the type specified by
                `context_types[i]`. Valid types include the following: `categorical`,
                `continuous`, `ordinal`, `count`, and `datetime`.
            data_types (list):
                List of strings indicating the type of each channel in data.
                Each value in the list at data[i] must match the type specified by
                `data_types[i]`. The valid types are the same as for `context_types`.
        """
        X, C = [], []
        self._build(sequences, context_types, data_types)
        for sequence in sequences:
            X.append(self._data_to_tensor(sequence['data']))
            C.append(self._context_to_tensor(sequence['context']))

        X = torch.nn.utils.rnn.pack_sequence(X, enforce_sorted=False).to(self.device)
        if self._ctx_dims:
            C = torch.stack(C, dim=0).to(self.device)

        self._model = PARNet(self._data_dims, self._ctx_dims).to(self.device)
        optimizer = torch.optim.Adam(self._model.parameters(), lr=1e-3)

        iterator = tqdm(range(self.epochs), disable=(not self.verbose))
        if self.verbose:
            pbar_description = 'Loss ({loss:.3f})'
            iterator.set_description(pbar_description.format(loss=0))

        # Reset loss_values dataframe
        self.loss_values = pd.DataFrame(columns=['Epoch', 'Loss'])

        X_padded, seq_len = torch.nn.utils.rnn.pad_packed_sequence(X)
        for epoch in iterator:
            Y = self._model(X, C)
            Y_padded, _ = torch.nn.utils.rnn.pad_packed_sequence(Y)

            optimizer.zero_grad()
            loss = self._compute_loss(X_padded[1:, :, :], Y_padded[:-1, :, :], seq_len)
            loss.backward()

            epoch_loss_df = pd.DataFrame({
                'Epoch': [epoch],
                'Loss': [loss.item()],
            })
            if not self.loss_values.empty:
                self.loss_values = pd.concat([
                    self.loss_values,
                    epoch_loss_df,
                ]).reset_index(drop=True)
            else:
                self.loss_values = epoch_loss_df

            if self.verbose:
                iterator.set_description(pbar_description.format(loss=loss.item()))

            optimizer.step()

    def _compute_loss(self, X_padded, Y_padded, seq_len):
        """Compute the loss between X and Y.

        Given X[i,:,:], the neural network predicts the value at the next
        timestep (i+1); this prediction is provided in Y[i,:,:]. This function
        returns the loss between the predicted and actual sequence.

        .. note::
            The `i`th time series (with padding removed) can be indexed with
            `X[:seq_len[i], i, :]`.

        Args:
            X_padded (tensor):
                This contains the input to the model.
            Y_padded (tensor):
                This contains the output of the model.
            seq_len (list):
                This list contains the length of each sequence.
        """
        log_likelihood = 0.0
        _, batch_size, _input_size = X_padded.shape

        for key, props in self._data_map.items():
            if props['type'] in ['continuous', 'timestamp']:
                mu_idx, sigma_idx, missing_idx = props['indices']
                mu = Y_padded[:, :, mu_idx]
                sigma = torch.nn.functional.softplus(Y_padded[:, :, sigma_idx])
                missing = torch.nn.LogSigmoid()(Y_padded[:, :, missing_idx])

                for i in range(batch_size):
                    dist = torch.distributions.normal.Normal(
                        mu[: seq_len[i], i], sigma[: seq_len[i], i]
                    )
                    log_likelihood += torch.sum(dist.log_prob(X_padded[-seq_len[i] :, i, mu_idx]))

                    p_true = X_padded[: seq_len[i], i, missing_idx]
                    p_pred = missing[: seq_len[i], i]
                    log_likelihood += torch.sum(p_true * p_pred)
                    log_likelihood += torch.sum((1.0 - p_true) * torch.log(1.0 - torch.exp(p_pred)))

            elif props['type'] in ['count']:
                r_idx, p_idx, missing_idx = props['indices']
                r = torch.nn.functional.softplus(Y_padded[:, :, r_idx]) * props['range']
                p = torch.sigmoid(Y_padded[:, :, p_idx])
                x = X_padded[:, :, r_idx] * props['range']
                missing = torch.nn.LogSigmoid()(Y_padded[:, :, missing_idx])

                for i in range(batch_size):
                    dist = torch.distributions.negative_binomial.NegativeBinomial(
                        r[: seq_len[i], i],
                        p[: seq_len[i], i],
                        validate_args=False,
                    )
                    log_likelihood += torch.sum(dist.log_prob(x[: seq_len[i], i]))

                    p_true = X_padded[: seq_len[i], i, missing_idx]
                    p_pred = missing[: seq_len[i], i]
                    log_likelihood += torch.sum(p_true * p_pred)
                    log_likelihood += torch.sum((1.0 - p_true) * torch.log(1.0 - torch.exp(p_pred)))

            elif props['type'] in ['categorical', 'ordinal']:
                idx = list(props['indices'].values())
                log_softmax = torch.nn.functional.log_softmax(Y_padded[:, :, idx], dim=2)

                for i in range(batch_size):
                    target = X_padded[: seq_len[i], i, idx]
                    predicted = log_softmax[: seq_len[i], i]
                    target = torch.argmax(target, dim=1).unsqueeze(dim=1)
                    log_likelihood += torch.sum(predicted.gather(dim=1, index=target))

            else:
                raise ValueError()

        return -log_likelihood / (batch_size * len(self._data_map) * batch_size)

    def _tensor_to_data(self, x):
        # Force CPU on x
        x = x.to(torch.device('cpu'))

        seq_len, batch_size, _ = x.shape
        assert batch_size == 1

        data = [None] * (len(self._data_map) - 1)
        for key, props in self._data_map.items():
            if key == '<TOKEN>':
                continue

            data[key] = []
            for i in range(seq_len):
                if props['type'] in ['continuous', 'datetime']:
                    mu_idx, _sigma_idx, missing_idx = props['indices']
                    if (x[i, 0, missing_idx] > 0) and props['nulls']:
                        data[key].append(None)
                    else:
                        data[key].append(x[i, 0, mu_idx].item() * props['std'] + props['mu'])

                elif props['type'] in ['count']:
                    r_idx, _p_idx, missing_idx = props['indices']
                    if x[i, 0, missing_idx] > 0 and props['nulls']:
                        data[key].append(None)
                    else:
                        sample = x[i, 0, r_idx].item() * props['range'] + props['min']
                        data[key].append(int(sample))

                elif props['type'] in ['categorical', 'ordinal']:
                    ml_value, max_x = None, float('-inf')
                    for value, idx in props['indices'].items():
                        if x[i, 0, idx] > max_x:
                            max_x = x[i, 0, idx]
                            ml_value = value

                    data[key].append(ml_value)

                else:
                    raise ValueError()

        return data

    def _sample_state(self, x):
        log_likelihood = 0.0
        seq_len, batch_size, _input_size = x.shape
        assert seq_len == 1 and batch_size == 1

        for key, props in self._data_map.items():
            if props['type'] in ['continuous', 'timestamp']:
                mu_idx, sigma_idx, missing_idx = props['indices']
                mu = x[0, 0, mu_idx]
                sigma = torch.nn.functional.softplus(x[0, 0, sigma_idx])
                dist = torch.distributions.normal.Normal(mu, sigma)
                x[0, 0, mu_idx] = dist.sample()
                x[0, 0, sigma_idx] = 0.0
                log_likelihood += torch.sum(dist.log_prob(x[0, 0, mu_idx]))

                dist = torch.distributions.Bernoulli(torch.sigmoid(x[0, 0, missing_idx]))
                x[0, 0, missing_idx] = dist.sample()
                x[0, 0, mu_idx] = x[0, 0, mu_idx] * (1.0 - x[0, 0, missing_idx])
                log_likelihood += torch.sum(dist.log_prob(x[0, 0, missing_idx]))

            elif props['type'] in ['count']:
                r_idx, p_idx, missing_idx = props['indices']
                r = torch.nn.functional.softplus(x[0, 0, r_idx]) * props['range']
                p = torch.sigmoid(x[0, 0, p_idx])
                dist = torch.distributions.negative_binomial.NegativeBinomial(r, p)
                x[0, 0, r_idx] = dist.sample()
                x[0, 0, p_idx] = 0.0
                log_likelihood += torch.sum(dist.log_prob(x[0, 0, r_idx]))
                x[0, 0, r_idx] /= props['range']

                dist = torch.distributions.Bernoulli(torch.sigmoid(x[0, 0, missing_idx]))
                x[0, 0, missing_idx] = dist.sample()
                x[0, 0, r_idx] = x[0, 0, r_idx] * (1.0 - x[0, 0, missing_idx])
                log_likelihood += torch.sum(dist.log_prob(x[0, 0, missing_idx]))

            elif props['type'] in ['categorical', 'ordinal']:
                idx = list(props['indices'].values())
                p = torch.nn.functional.softmax(x[0, 0, idx], dim=0)
                x_new = torch.zeros(p.size()).to(self.device)
                x_new.scatter_(dim=0, index=torch.multinomial(p, 1), value=1)
                x[0, 0, idx] = x_new
                log_likelihood += torch.sum(torch.log(p) * x_new)

            else:
                raise ValueError()

        return x, log_likelihood

    def _sample_sequence(self, context, min_length, max_length):
        log_likelihood = 0.0

        x = torch.zeros(self._data_dims).to(self.device)
        x[self._data_map['<TOKEN>']['indices']['<START>']] = 1.0
        x = x.unsqueeze(0).unsqueeze(0)

        for step in range(max_length):
            next_x, ll = self._sample_state(self._model(x, context)[-1:, :, :])
            x = torch.cat([x, next_x], dim=0)
            log_likelihood += ll
            if next_x[0, 0, self._data_map['<TOKEN>']['indices']['<END>']] > 0.0:
                if min_length <= step + 1 <= max_length:
                    break  # received end token

                next_x[0, 0, self._data_map['<TOKEN>']['indices']['<BODY>']] = 1.0
                next_x[0, 0, self._data_map['<TOKEN>']['indices']['<END>']] = 0.0

        return x[1:, :, :], log_likelihood

    def sample_sequence(self, context, sequence_length=None):
        """Sample a single sequence conditioned on context.

        Args:
            context (list):
                The list of values to condition on. It must match
                the types specified in context_types when fit was called.
            sequence_length (int or None):
                If given, force sequences to be of the indicated length.
                If ``None`` (default), sample sequences of the same length
                as the original dataset.

        Returns:
            list[list]:
                A list of lists (data) corresponding to the types specified
                in data_types when fit was called.
        """
        if sequence_length is not None:
            min_length = max_length = sequence_length
        else:
            min_length = self._min_length
            max_length = self._max_length

        if self._ctx_dims:
            context = self._context_to_tensor(context).unsqueeze(0)
        else:
            context = None

        best_x, best_ll = None, float('-inf')
        for _ in range(self.sample_size):
            with torch.no_grad():
                x, log_likelihood = self._sample_sequence(context, min_length, max_length)

            if log_likelihood > best_ll:
                best_x = x
                best_ll = log_likelihood

        return self._tensor_to_data(best_x)



"""Tests for the ``tasks.py`` file."""

from tasks import _get_minimum_versions


def test_get_minimum_versions():
    """Test the ``_get_minimum_versions`` method.

    The method should return the minimum versions of the dependencies for the given python version.
    If a library is linked to an URL, the minimum version should be the URL.
    """
    # Setup
    dependencies = [
        "numpy>=1.20.0,<2;python_version<'3.10'",
        "numpy>=1.23.3,<2;python_version>='3.10'",
        "pandas>=1.2.0,<2;python_version<'3.10'",
        "pandas>=1.3.0,<2;python_version>='3.10'",
        'humanfriendly>=8.2,<11',
        'pandas @ git+https://github.com/pandas-dev/pandas.git@master',
    ]

    # Run
    minimum_versions_39 = _get_minimum_versions(dependencies, '3.9')
    minimum_versions_310 = _get_minimum_versions(dependencies, '3.10')

    # Assert
    expected_versions_39 = [
        'numpy==1.20.0',
        'git+https://github.com/pandas-dev/pandas.git@master#egg=pandas',
        'humanfriendly==8.2',
    ]
    expected_versions_310 = [
        'numpy==1.23.3',
        'git+https://github.com/pandas-dev/pandas.git@master#egg=pandas',
        'humanfriendly==8.2',
    ]

    assert minimum_versions_39 == expected_versions_39
    assert minimum_versions_310 == expected_versions_310



"""Base DeepEcho class."""

import pandas as pd
import tqdm

from deepecho.sequences import assemble_sequences


class DeepEcho:
    """The base class for DeepEcho models."""

    _verbose = True
    _output_columns = None
    _data_columns = None
    _entity_columns = None
    _context_columns = None
    _context_values = None

    @staticmethod
    def _validate(sequences, context_types, data_types):
        """Validate the model input.

        Args:
            sequences:
                See `fit`.
            context_types:
                See `fit`.
            data_types:
                See `fit`.
        """
        dtypes = set([
            'continuous',
            'categorical',
            'ordinal',
            'count',
            'datetime',
        ])
        assert all(dtype in dtypes for dtype in context_types)
        assert all(dtype in dtypes for dtype in data_types)

        for sequence in sequences:
            assert len(sequence['context']) == len(context_types)
            assert len(sequence['data']) == len(data_types)
            lengths = [len(x) for x in sequence['data']]
            assert len(set(lengths)) == 1

    def fit_sequences(self, sequences, context_types, data_types):
        """Fit a model to the specified sequences.

        Args:
            sequences:
                List of sequences. Each sequence is a single training example
                (i.e. an example of a multivariate time series with some context).
                For example, a sequence might look something like::

                    {
                        'context': [1],
                        'data': [
                            [1, 3, 4, 5, 11, 3, 4],
                            [2, 2, 3, 4, 5, 1, 2],
                            [1, 3, 4, 5, 2, 3, 1],
                        ],
                    }

                The "context" attribute maps to a list of variables which
                should be used for conditioning. These are variables which
                do not change over time.

                The "data" attribute contains a list of lists corrsponding
                to the actual time series data such that `data[i][j]` contains
                the value at the jth time step of the ith channel of the
                multivariate time series.
            context_types:
                List of strings indicating the type of each value in context.
                he value at `context[i]` must match the type specified by
                `context_types[i]`. Valid types include the following: `categorical`,
                `continuous`, `ordinal`, `count`, and `datetime`.
            data_types:
                List of strings indicating the type of each channel in data.
                Each value in the list at data[i] must match the type specified by
                `data_types[i]`. The valid types are the same as for `context_types`.
        """
        raise NotImplementedError()

    @staticmethod
    def _get_data_types(data, data_types, columns):
        """Analyze the data and tell the data type of each column."""
        dtypes_list = []
        data_types = data_types or {}
        for column in columns:
            if column in data_types:
                dtypes_list.append(data_types[column])
            else:
                dtype = data[column].dtype
                kind = dtype.kind
                if kind in 'fiud':
                    dtypes_list.append('continuous')
                elif kind in 'OSUb':
                    dtypes_list.append('categorical')
                elif kind == 'M':
                    dtypes_list.append('datetime')
                else:
                    error = f'Unsupported data_type for column {column}: {dtype}'
                    raise ValueError(error)

        return dtypes_list

    def fit(
        self,
        data,
        entity_columns=None,
        context_columns=None,
        data_types=None,
        segment_size=None,
        sequence_index=None,
    ):
        """Fit the model to a dataframe containing time series data.

        Args:
            data (pd.DataFrame):
                DataFrame containing all the timeseries data alongside the
                entity and context columns.
            entity_columns (list[str]):
                Names of the columns which identify different time series
                sequences. These will be used to group the data in separated
                training examples.
            context_columns (list[str]):
                The columns in the dataframe which are constant within each
                group/entity. These columns will be provided at sampling time
                (i.e. the samples will be conditioned on the context variables).
            data_types (dict[str, str]):
                Dictinary indicating the data types of each column, which can be
                ``categorical``, ``continuous`` or ``datetime``.
            segment_size (int, pd.Timedelta or str):
                If specified, cut each training sequence in several segments of the
                indicated size. The size can either can passed as an integer value,
                which will interpreted as the number of data points to put on each
                segment, or as a pd.Timedelta (or equivalent str representation),
                which will be interpreted as the segment length in time. Timedelta
                segment sizes can only be used with sequence indexes of type datetime.
            sequence_index (str):
                Name of the column that acts as the order index of each sequence.
                The sequence index column can be of any type that can be sorted,
                such as integer values or datetimes.
        """
        if not entity_columns and segment_size is None:
            raise TypeError('If the data has no `entity_columns`, `segment_size` must be given.')
        if segment_size is not None and not isinstance(segment_size, int):
            if sequence_index is None:
                raise TypeError(
                    '`segment_size` must be of type `int` if no `sequence_index` is given.'
                )
            if data[sequence_index].dtype.kind != 'M':
                raise TypeError(
                    '`segment_size` must be of type `int` if '
                    '`sequence_index` is not a `datetime` column.'
                )

            segment_size = pd.to_timedelta(segment_size)

        self._output_columns = list(data.columns)
        self._entity_columns = entity_columns or []
        self._context_columns = context_columns or []
        self._data_columns = [
            column
            for column in data.columns
            if column not in self._entity_columns + self._context_columns
        ]
        if sequence_index:
            self._output_columns.remove(sequence_index)
            self._data_columns.remove(sequence_index)

        data_types = self._get_data_types(data, data_types, self._data_columns)
        context_types = self._get_data_types(data, data_types, self._context_columns)
        sequences = assemble_sequences(
            data,
            self._entity_columns,
            self._context_columns,
            segment_size,
            sequence_index,
        )

        # Validate and fit
        self._validate(sequences, context_types, data_types)
        self.fit_sequences(sequences, context_types, data_types)

        # Store context values
        self._context_values = data[self._context_columns]

    def sample_sequence(self, context, sequence_length=None):
        """Sample a single sequence conditioned on context.

        Args:
            context (list):
                The list of values to condition on. It must match
                the types specified in context_types when fit was called.
            sequence_length (int or None):
                If given, force sequences to be of the indicated length.
                If ``None`` (default), sample sequences of the same length
                as the original dataset.

        Returns:
            list[list]:
                A list of lists (data) corresponding to the types specified
                in data_types when fit was called.
        """
        raise NotImplementedError()

    def sample(self, num_entities=None, context=None, sequence_length=None):
        """Sample a dataframe containing time series data.

        Args:
            num_entities (int):
                The number of entities to sample.
            context (pd.DataFrame):
                Context values to use when sampling.
            sequence_length (int or None):
                If given, force sequences to be of the indicated length.
                If ``None`` (default), sample sequences of the same length
                as the original dataset.

        Returns:
            pd.DataFrame:
                A DataFrame which resembles the original dataframe where (1) the
                entity column(s) are arbitrarily generated, (2) the context
                column(s) are resampled from the original data, and (3) the data
                columns containing the time series comes from the conditional
                time series model.
        """
        if context is None:
            if num_entities is None:
                raise TypeError('Either context or num_entities must be not None')

            context = self._context_values.sample(num_entities, replace=True)
            context = context.reset_index(drop=True)

        else:
            num_entities = len(context)
            context = context.copy()

        for column in self._entity_columns:
            if column not in context:
                context[column] = range(num_entities)

        # Set the entity_columns as index to properly iterate over them
        if self._entity_columns:
            context = context.set_index(self._entity_columns)

        if self._verbose:
            iterator = tqdm.tqdm(context.iterrows(), total=num_entities)
        else:
            iterator = context.iterrows()

        output = pd.DataFrame()
        for entity_values, context_values in iterator:
            context_values = context_values.tolist()
            sequence = self.sample_sequence(context_values, sequence_length)

            # Reformat as a DataFrame
            group = pd.DataFrame(
                dict(zip(self._data_columns, sequence)),
                columns=self._data_columns,
            )
            group[self._entity_columns] = entity_values
            for column, value in zip(self._context_columns, context_values):
                group[column] = value

            output = pd.concat([output, group])

        return output[self._output_columns].reset_index(drop=True)



"""Top-level package for DeepEcho."""

__author__ = 'DataCebo, Inc.'
__email__ = 'info@sdv.dev'
__version__ = '0.7.0'
__path__ = __import__('pkgutil').extend_path(__path__, __name__)

from deepecho.demo import load_demo
from deepecho.models.basic_gan import BasicGANModel
from deepecho.models.par import PARModel

__all__ = [
    'load_demo',
    'BasicGANModel',
    'PARModel',
]



"""Functions to load demo data."""

import os

import pandas as pd

_DATA_PATH = os.path.join(os.path.dirname(__file__), 'data')


def load_demo():
    """Load the demo DataFrame."""
    return pd.read_csv(os.path.join(_DATA_PATH, 'demo.csv'), parse_dates=['date'])



"""BasicGAN Model."""

import logging

import numpy as np
import pandas as pd
import torch
from tqdm import tqdm

from deepecho.models.base import DeepEcho

LOGGER = logging.getLogger(__name__)


def _expand_context(data, context):
    return torch.cat(
        [
            data,
            context.unsqueeze(0).expand(data.shape[0], context.shape[0], context.shape[1]),
        ],
        dim=2,
    )


class BasicGenerator(torch.nn.Module):
    """Generator for the BasicGAN model.

    This generator consist on a RNN layer followed by a Linear layer with
    the following schema:

        - The Generator takes as input a ``sequence_length`` and a ``context`` vector.
        - The ``context`` vector is expanded over the ``sequence_lenght`` and padded with
          ``latent_size`` random noise.
        - RNN takes as input a tensor with shape
          ``(sequence_length, context_length, context_size + latent_size)`` and
          generates an output of shape ``(sequence_length, context_length, hidden_size)``.
        - The RNN output is passed to the Linear layer that outputs a tensor of size
          ``(sequence_length, context_length, output_size)``

    Args:
        context_size (int):
            Size of the contextual arrays.
        latent_size (int):
            Size of the random input vector.
        hidden_size (int):
            Size of the communication between the RNN and the Linear layer.
        data_size (int):
            Size of the output layer.
        device (torch.device):
            Device to which this Module is associated to.
    """

    def __init__(self, context_size, latent_size, hidden_size, data_size, device):
        super().__init__()
        self.latent_size = latent_size
        self.rnn = torch.nn.GRU(context_size + latent_size, hidden_size)
        self.linear = torch.nn.Linear(hidden_size, data_size)
        self.device = device

    def forward(self, context=None, sequence_length=None):
        """Forward computation.

        Args:
            context (tensor):
                Context values to use for each generated sequence.
            sequence_length (int):
                Amount of data points to generate for each sequence.
        """
        latent = torch.randn(
            size=(sequence_length, context.size(0), self.latent_size),
            device=self.device,
        )
        latent = _expand_context(latent, context)

        rnn_out, _ = self.rnn(latent)
        return self.linear(rnn_out)


class BasicDiscriminator(torch.nn.Module):
    """Discriminator for the BasicGAN model.

    This discriminator consist on a RNN layer followed by a Linear layer with
    the following schema:

        - The Discriminator takes as input a collection of sequences that include
          both the data and the context columns.
        - RNN takes as input a tensor with shape
          ``(sequence_length, number_of_sequences, context_size + data_size)`` and
          generates an output of shape ``(sequence_length, num_sequences, hidden_size)``.
        - The RNN output is passed to the Linear layer that outputs a single value.

    Args:
        context_size (int):
            Number of values in the contextual arrays.
        data_size (int):
            Number of variables in the input data sequences.
        hidden_size (int):
            Size of the communication between the RNN and the Linear layer.
    """

    def __init__(self, context_size, data_size, hidden_size):
        super().__init__()
        self.rnn = torch.nn.GRU(context_size + data_size, hidden_size)
        self.linear = torch.nn.Linear(hidden_size, 1)

    def forward(self, sequences):
        """Forward computation.

        Args:
            sequences (tensor):
                Sequences of values with their context.
        """
        rnn_out, _ = self.rnn(sequences)
        return self.linear(rnn_out[-1, :, :])


class BasicGANModel(DeepEcho):
    """Basic GAN model.

    0. Normalize continuous values to [-1.0, 1.0].
    1. Map categorical values to one-hot.
    2. Define a generator that takes context vector -> new sequence.
    3. Transforms
        - apply sigmoid to continuous/count/datetime
        - apply softmax to categorical/ordinal
    4. Define a discriminator that takes sequence + context -> score.

    Args:
        epochs (int):
            Number of training epochs. Defaults to 1024.
        latent_size (int):
            Size of the random vector to use for generation. Defaults to 32.
        hidden_size (int):
            Size of the hidden Linear layer. Defaults to 16.
        gen_lr (float):
            Generator learning rate. Defaults to 1e-3.
        dis_lr (float):
            Discriminator learning rate. Defaults to 1e-3.
        cuda (bool):
            Whether to attempt to use cuda for GPU computation.
            If this is False or CUDA is not available, CPU will be used.
            Defaults to ``True``.
        verbose (bool):
            Whether to print progress to console or not.
    """

    _max_sequence_length = None
    _fixed_length = None
    _context_map = None
    _context_size = None
    _data_map = None
    _data_size = None
    _model_data_size = None
    _generator = None

    def __init__(
        self,
        epochs=1024,
        latent_size=32,
        hidden_size=16,
        gen_lr=1e-3,
        dis_lr=1e-3,
        cuda=True,
        verbose=True,
    ):
        self._epochs = epochs
        self._gen_lr = gen_lr
        self._dis_lr = dis_lr
        self._latent_size = latent_size
        self._hidden_size = hidden_size

        if not cuda or not torch.cuda.is_available():
            device = 'cpu'
        elif isinstance(cuda, str):
            device = cuda
        else:
            device = 'cuda'

        self._device = torch.device(device)
        self._verbose = verbose

        LOGGER.info('%s instance created', self)

    def __repr__(self):
        """Return a representation of the class object."""
        return (
            f'{self.__class__.__name__}(epochs={self._epochs}, latent_size={self._latent_size},'
            f'hidden_size={self._hidden_size}, gen_lr={self._gen_lr}, dis_lr={self._dis_lr},'
            f"cuda='{self._device}', verbose={self._verbose})"
        )

    # ########################### #
    # Preprocessing and preparing #
    # ########################### #

    @staticmethod
    def _index_map(columns, types):
        """Decide which dimension will store which column information in the tensor.

        The output of this function has two elements:

            - An idx_map, which is a dict that indicates the indexes at which
              the list of tensor dimensions associated with each input column starts,
              and the properties of such columns.
            - An integer that indicates how many dimensions the tensor will have.

        In order to decide this, the following process is followed for each column:

            - If the column is numerical (continuous or count), 2 dimensions are created
              for it. These will contain information about the value itself, as well
              as information about whether the value should be NaN or not.
            - If the column is categorical or ordinal, 1 dimentions is created for
              each possible value, which will be later on used to hold one-hot encoding
              information about the values.
        """
        dimensions = 0
        mapping = {}
        for column, column_type in enumerate(types):
            values = columns[column]
            if column_type in ('continuous', 'count'):
                mapping[column] = {
                    'type': column_type,
                    'min': np.min(values),
                    'max': np.max(values),
                    'indices': (dimensions, dimensions + 1),
                }
                dimensions += 2

            elif column_type in ('categorical', 'ordinal'):
                indices = {}
                for value in set(values):
                    indices[value] = dimensions
                    dimensions += 1

                mapping[column] = {'type': column_type, 'indices': indices}

            else:
                raise ValueError(f'Unsupported type: {column_type}')

        return mapping, dimensions

    def _analyze_data(self, sequences, context_types, data_types):
        """Extract information about the context and data that will be used later.

        The following information is stored:
            - Length of the sequences.
            - Index map and dimensions for the context.
            - Index map and dimensions for the data.
        """
        sequence_lengths = np.array([len(sequence['data'][0]) for sequence in sequences])
        self._max_sequence_length = np.max(sequence_lengths)
        self._fixed_length = (sequence_lengths == self._max_sequence_length).all()

        # Concatenate all the context sequences together
        context = []
        for column in range(len(context_types)):
            context.append([sequence['context'][column] for sequence in sequences])

        self._context_map, self._context_size = self._index_map(context, context_types)

        # Concatenate all the data sequences together
        data = []
        for column in range(len(data_types)):
            data.append(sum([sequence['data'][column] for sequence in sequences], []))

        self._data_map, self._data_size = self._index_map(data, data_types)

        self._model_data_size = self._data_size + int(not self._fixed_length)

    @staticmethod
    def _normalize(tensor, value, properties):
        """Normalize the value between 0 and 1 and flag nans."""
        value_idx, missing_idx = properties['indices']
        if pd.isna(value):
            tensor[value_idx] = 0.0
            tensor[missing_idx] = 1.0
        else:
            column_min = properties['min']
            column_range = properties['max'] - column_min
            offset = value - column_min
            tensor[value_idx] = 2.0 * offset / column_range - 1.0
            tensor[missing_idx] = 0.0

    @staticmethod
    def _denormalize(tensor, row, properties, round_value):
        """Denormalize previously normalized values, setting NaN values if necessary."""
        value_idx, missing_idx = properties['indices']
        if tensor[row, 0, missing_idx] > 0.5:
            return None

        normalized = tensor[row, 0, value_idx].item()
        column_min = properties['min']
        column_range = properties['max'] - column_min

        denormalized = (normalized + 1) * column_range / 2.0 + column_min
        if round_value:
            denormalized = round(denormalized)

        return denormalized

    @staticmethod
    def _one_hot_encode(tensor, value, properties):
        """Update the index that corresponds to the value to 1.0."""
        value_index = properties['indices'][value]
        tensor[value_index] = 1.0

    @staticmethod
    def _one_hot_decode(tensor, row, properties):
        """Obtain the category that corresponds to the highest one-hot value."""
        max_value = float('-inf')
        for category, idx in properties['indices'].items():
            value = tensor[row, 0, idx]
            if value > max_value:
                max_value = value
                selected = category

        return selected

    def _value_to_tensor(self, tensor, value, properties):
        """Update the tensor according to the value and properties."""
        column_type = properties['type']
        if column_type in ('continuous', 'count'):
            self._normalize(tensor, value, properties)
        elif column_type in ('categorical', 'ordinal'):
            self._one_hot_encode(tensor, value, properties)

        else:
            raise ValueError()  # Theoretically unreachable

    def _data_to_tensor(self, data):
        """Convert the input data to the corresponding tensor.

        If ``self._fixed_length`` is ``False``, add a 1.0 to indicate
        the sequence end and pad the rest of the sequence with 0.0s.
        """
        tensors = []
        num_rows = len(data[0])
        for row in range(num_rows):
            tensor = torch.zeros(self._model_data_size)
            for column, properties in self._data_map.items():
                value = data[column][row]
                self._value_to_tensor(tensor, value, properties)

            tensors.append(tensor)

        if not self._fixed_length:
            tensors[-1][-1] = 1.0

        for _ in range(self._max_sequence_length - num_rows):
            tensors.append(torch.zeros(self._model_data_size))

        return torch.stack(tensors, dim=0)

    def _context_to_tensor(self, context):
        """Convert the input context to the corresponding tensor."""
        tensor = torch.zeros(self._context_size)
        for column, properties in self._context_map.items():
            value = context[column]
            self._value_to_tensor(tensor, value, properties)

        return tensor

    def _tensor_to_data(self, tensor):
        """Rebuild a valid sequence from the given tensor."""
        sequence_length, num_sequences, _ = tensor.shape
        assert num_sequences == 1

        data = [None] * len(self._data_map)
        for column, properties in self._data_map.items():
            column_type = properties['type']

            column_data = []
            data[column] = column_data
            for row in range(sequence_length):
                if column_type in ('continuous', 'count'):
                    round_value = column_type == 'count'
                    value = self._denormalize(tensor, row, properties, round_value=round_value)
                elif column_type in ('categorical', 'ordinal'):
                    value = self._one_hot_decode(tensor, row, properties)
                else:
                    raise ValueError()  # Theoretically unreachable

                column_data.append(value)

        return data

    def _build_tensor(self, transform, sequences, key, dim):
        """Convert input sequences to tensors."""
        tensors = []
        for sequence in sequences:
            tensors.append(transform(sequence[key]))

        return torch.stack(tensors, dim=dim).to(self._device)

    # ################## #
    # GAN Training steps #
    # ################## #

    def _transform(self, data):
        for properties in self._data_map.values():
            column_type = properties['type']
            if column_type in ('continuous', 'count'):
                value_idx, missing_idx = properties['indices']
                data[:, :, value_idx] = torch.tanh(data[:, :, value_idx])
                data[:, :, missing_idx] = torch.sigmoid(data[:, :, missing_idx])
            elif column_type in ('categorical', 'ordinal'):
                indices = list(properties['indices'].values())
                data[:, :, indices] = torch.nn.functional.softmax(data[:, :, indices])

        return data

    def _truncate(self, generated):
        end_flag = (generated[:, :, self._data_size] > 0.5).float()
        end_flag = end_flag.round()
        generated[:, :, self._data_size] = end_flag

        for sequence_idx in range(generated.shape[1]):
            # Pad with zeroes after end_flag == 1
            sequence = generated[:, sequence_idx]
            end_flag = sequence[:, self._data_size]
            if (end_flag == 1.0).any():
                cut_idx = end_flag.detach().cpu().numpy().argmax()
                sequence[cut_idx + 1 :] = 0.0

    def _generate(self, context, sequence_length=None):
        generated = self._generator(
            context=context,
            sequence_length=sequence_length or self._max_sequence_length,
        )

        generated = self._transform(generated)
        if not self._fixed_length:
            self._truncate(generated)

        return generated

    def _discriminator_step(self, discriminator, discriminator_opt, data_context, context):
        real_scores = discriminator(data_context)

        fake = self._generate(context)
        fake_context = _expand_context(fake, context)
        fake_scores = discriminator(fake_context)

        discriminator_score = -torch.mean(real_scores - fake_scores)

        discriminator_opt.zero_grad()
        discriminator_score.backward()
        discriminator_opt.step()

        for parameter in discriminator.parameters():
            parameter.data.clamp_(-0.01, 0.01)

        return discriminator_score

    def _generator_step(self, discriminator, generator_opt, context):
        fake = self._generate(context)
        fake_context = _expand_context(fake, context)
        generator_score = -torch.mean(discriminator(fake_context))

        generator_opt.zero_grad()
        generator_score.backward()
        generator_opt.step()

        return generator_score

    def _build_fit_artifacts(self):
        self._generator = BasicGenerator(
            context_size=self._context_size,
            latent_size=self._latent_size,
            hidden_size=self._hidden_size,
            data_size=self._model_data_size,
            device=self._device,
        ).to(self._device)

        discriminator = BasicDiscriminator(
            context_size=self._context_size,
            data_size=self._model_data_size,
            hidden_size=self._hidden_size,
        ).to(self._device)

        generator_opt = torch.optim.Adam(self._generator.parameters(), lr=self._gen_lr)
        discriminator_opt = torch.optim.Adam(discriminator.parameters(), lr=self._dis_lr)

        return discriminator, generator_opt, discriminator_opt

    def fit_sequences(self, sequences, context_types, data_types):
        """Fit a model to the specified sequences.

        Args:
            sequences:
                List of sequences. Each sequence is a single training example
                (i.e. an example of a multivariate time series with some context).
                For example, a sequence might look something like::

                    {
                        'context': [1],
                        'data': [
                            [1, 3, 4, 5, 11, 3, 4],
                            [2, 2, 3, 4, 5, 1, 2],
                            [1, 3, 4, 5, 2, 3, 1],
                        ],
                    }

                The "context" attribute maps to a list of variables which
                should be used for conditioning. These are variables which
                do not change over time.

                The "data" attribute contains a list of lists corrsponding
                to the actual time series data such that `data[i][j]` contains
                the value at the jth time step of the ith channel of the
                multivariate time series.
            context_types:
                List of strings indicating the type of each value in context.
                he value at `context[i]` must match the type specified by
                `context_types[i]`. Valid types include the following: `categorical`,
                `continuous`, `ordinal`, `count`, and `datetime`.
            data_types:
                List of strings indicating the type of each channel in data.
                Each value in the list at data[i] must match the type specified by
                `data_types[i]`. The valid types are the same as for `context_types`.
        """
        self._analyze_data(sequences, context_types, data_types)

        data = self._build_tensor(self._data_to_tensor, sequences, 'data', dim=1)
        context = self._build_tensor(self._context_to_tensor, sequences, 'context', dim=0)
        data_context = _expand_context(data, context)

        discriminator, generator_opt, discriminator_opt = self._build_fit_artifacts()

        iterator = range(self._epochs)
        if self._verbose:
            iterator = tqdm(iterator)

        for epoch in iterator:
            discriminator_score = self._discriminator_step(
                discriminator=discriminator,
                discriminator_opt=discriminator_opt,
                data_context=data_context,
                context=context,
            )
            generator_score = self._generator_step(
                discriminator=discriminator,
                generator_opt=generator_opt,
                context=context,
            )

            if self._verbose:
                d_loss = discriminator_score.item()
                g_loss = generator_score.item()
                iterator.set_description(f'Epoch {epoch + 1} | D Loss {d_loss} | G Loss {g_loss}')

    def sample_sequence(self, context, sequence_length=None):
        """Sample a single sequence conditioned on context.

        Args:
            context (list):
                The list of values to condition on. It must match
                the types specified in context_types when fit was called.

        Returns:
            list[list]:
                A list of lists (data) corresponding to the types specified
                in data_types when fit was called.
        """
        context = self._context_to_tensor(context).unsqueeze(0).to(self._device)

        with torch.no_grad():
            generated = self._generate(context, sequence_length)
            if sequence_length is None:
                end_flag = generated[:, 0, -1]
                if (end_flag == 1.0).any():
                    cut_index = end_flag.cpu().numpy().argmax()
                    generated = generated[:cut_index, :, :]

            return self._tensor_to_data(generated)



"""Functions to manipulate sequences and assemble training examples."""


def segment_by_size(sequence, segment_size):
    """Segment the sequence in segments of the indicated size.

    If sequence length is not exactly divisible by the ``segment_size``,
    extra data points at the end will be discarded to ensure that all
    the segments have the same size.

    Args:
        sequence (pandas.DataFrame):
            Sequence to segment, passed as a multi-column ``pandas.DataFrame``.
        segment_size (int):
            Size of each segment, passed as an integer.

    Returns:
        list:
            List of ``pandas.DataFrames`` containing each segment, all
            of the indicated size.
    """
    sequences = []
    start = 0
    total = len(sequence)
    while start < total:
        end = start + segment_size
        segment = sequence.iloc[start:end].reset_index(drop=True)
        if len(segment) != segment_size:
            break

        sequences.append(segment)
        start = end

    return sequences


def segment_by_time(sequence, segment_size, sequence_index):
    """Segment the sequence in segments of the indicated time length.

    Segmentation will happen by time, which means that there is no guarantee
    that the outputed segments all contain the same number of data points.

    Args:
        sequence (pandas.DataFrame):
            Sequence to segment, passed as a multi-column ``pandas.DataFrame``.
        segment_size (pandas.Timedelta):
            Size of each segment, passed as a ``pandas.Timedelta`` object.
        sequence_index (pandas.Series):
            Data of the column that will be used as the time index for the
            segmentation.

    Returns:
        list:
            List of ``pandas.DataFrames`` containing each segment.
    """
    sequences = []
    start = sequence_index.iloc[0]
    max_time = sequence_index.iloc[-1]
    while start <= max_time:
        end = start + segment_size
        selected = (start <= sequence_index) & (sequence_index < end)
        sequences.append(sequence[selected.to_numpy()].reset_index(drop=True))
        start = end

    return sequences


def segment_sequence(sequence, segment_size, sequence_index, drop_sequence_index=True):
    """Segment the sequence in segments of the indicated time length or size.

    If a ``sequence_index`` is given, data will be sorted by it first.

    If ``segment_size`` is not passed, the whole sequence will be returned
    as a single segment.

    Args:
        sequence (pandas.DataFrame):
            Sequence to segment, passed as a multi-column ``pandas.DataFrame``.
        segment_size (int or pandas.Timedelta):
            Size of each segment, passed as an integer or as a``pandas.Timedelta``
            object.
        sequence_index (str):
            Name of the column that will be used as the time index for the
            segmentation. Required if a timedelta ``segment_size`` is passed.
        drop_sequence_index (bool):
            Whether to drop the sequence index after sorting. Defaults to ``True``.

    Returns:
        list:
            List of ``pandas.DataFrames`` containing each segment.
    """
    if sequence_index is not None:
        sequence = sequence.sort_values(sequence_index)
        sequence_index_values = sequence[sequence_index]
        if drop_sequence_index:
            del sequence[sequence_index]

    if segment_size is None:
        return [sequence]

    if isinstance(segment_size, int):
        return segment_by_size(sequence, segment_size)

    return segment_by_time(sequence, segment_size, sequence_index_values)


def _convert_to_dicts(segments, context_columns):
    sequences = []
    for segment in segments:
        if context_columns:
            context = segment[context_columns]
            if len(context.drop_duplicates()) > 1:
                raise ValueError('Context columns are not constant within each segment.')

            context = context.iloc[0].to_numpy()
            segment = segment.drop(context_columns, axis=1)
        else:
            context = []

        lists = [list(row) for _, row in segment.items()]
        sequences.append({'context': context, 'data': lists})

    return sequences


def assemble_sequences(
    data,
    entity_columns,
    context_columns,
    segment_size,
    sequence_index,
    drop_sequence_index=True,
):
    """Build sequences from the data, grouping first by entity and then segmenting by size.

    Input is a ``pandas.DataFrame`` containing all the data, lists of entity and context
    columns and instructions about how to segment each entity sequence.

    The process of building the sequences consists on:

    1. First group by the entity columns. If no entity columns are given this
       step is skipped.
    2. Then segment the data that corresponds to each data using the segment_size
       and sequence_index details.
    3. Finally build dictionaries out of each segment, containing two elements:
        * `context`: List of contextual values.
        * `data`: List containing one list for each data column, containing its values.

    Args:
        data (pandas.DataFrame):
            Data to assemble in sequences, containing entity columns, context columns
            and data columns.
        entity_columns (list):
            List with the names of the columns that form each entity_id.
        context_columns (list):
            List with the names of the columns that act as context for each entity or
            segment.
            Context values must be constant within each entity, if entity_columns are
            given, or within each segment otherwise.
        segment_size (int or pandas.Timedelta):
            Size of each segment, passed as an integer or as a``pandas.Timedelta``
            object.
        sequence_index (str):
            Name of the column that will be used as the time index for the
            segmentation. Required if a timedelta ``segment_size`` is passed.
        drop_sequence_index (bool):
            Whether to drop the sequence index after sorting. Defaults to ``True``.

    Raises:
        ValueError:
            If context columns are not constant within each entity or segment.

    Returns:
        list:
            List of ``pandas.DataFrames`` containing each segment.
    """
    if not entity_columns:
        segments = segment_sequence(data, segment_size, sequence_index, drop_sequence_index)
    else:
        segments = []
        groupby_columns = entity_columns[0] if len(entity_columns) == 1 else entity_columns
        for _, sequence in data.groupby(groupby_columns):
            sequence = sequence.drop(entity_columns, axis=1)
            if context_columns:
                if len(sequence[context_columns].drop_duplicates()) > 1:
                    raise ValueError('Context columns are not constant within each entity.')

            entity_segments = segment_sequence(
                sequence, segment_size, sequence_index, drop_sequence_index
            )
            segments.extend(entity_segments)

    return _convert_to_dicts(segments, context_columns)
